package poly;

/*
 * a pragmatic compiler.
 * (c:) 2001, Sassa for ZF#4
 */

import java.io.*;
import java.util.Random;

public class Poly{
	public final static int MAX_TARGETS=10;	// for a while. I could leave it to be 7, as many as register
	public final static int MAX_RESOURCES=15;	// for a while
						// just as many, as there could maximum be
	public final static int STACK_LEN=32000;	// he-he... ;-)
	public final static int DEFAULT_LINES_TO_DO=10;

	public static Rc resources[]=new Rc[MAX_RESOURCES];
	public static int rc_count=0;	// amount of resources;
	public static int rc_read=0;	// amount of readable rcs
	public static int rc_write=0;	// amount of writeable rcs

	public static Target targets[]=new Target[MAX_TARGETS];
	public static int tg_count=0;

	public static Random rnd=new Random();

	public static PrintStream out=System.out;

	public static int lines_to_do=DEFAULT_LINES_TO_DO;
	public static int lines_can_do=0;
	public static boolean force_endurance=false;	// forces Targets to endure till they will get 0 lines to do
							// then they should execute their minimal sequence possible

	public final static String helpscreen=
"(c) 2001 by Sassa. BLENDER: Pragmatic polymorphic generator\n"+
"usage: Blender [-f <filename>] [-s {<rc> <value>}] [-d {<rc> <value>}] [-l <#>]\n"+
"\t-f specifies a filename for output;\n\t\tconsole by default\n"+
	//"\t-c"+  /// i am not creating resources so far
"\t-s specifies a set of starting values for resrouces;\n\t\tzeros for all by default\n"+
"\t-d specifies a set of destination values for the resources;\n\t\tno targets by default\n"+
	//"\t-u"+ /// i am not specifying a list of resources that can be used
"\t-l specifies a number of lines to produce; 10 by default.Produces\n\t\tthe minimal "+
"possible number of lines, if cannot fit into that\n\n"+
"";

	public static void main(String [] args){
		final int HELP=0;
		final int FILE_NAME=1;
		final int SOURCE_VALUES=2;
		final int DESTINATION_VALUES=3;
		final int CODE_LINES=4;

	   try{
		Register ax = new Register("ax", 0);
		Register bx = new Register("bx", 0);
		Register cx = new Register("cx", 0);
		Register dx = new Register("dx", 0);
		Register si = new Register("si", 0);
		Register di = new Register("di", 0);
		Register bp = new Register("bp", 0);
		new Stack("", STACK_LEN);


		int parser_state=HELP;
		Rc r;

		for (int i=0; i<args.length; i++){
			if (args[i].equals("-f")){
				parser_state=FILE_NAME;
				continue;
			}

			if (args[i].equals("-s")){
				parser_state=SOURCE_VALUES;
				continue;
			}

			if (args[i].equals("-d")){
				parser_state=DESTINATION_VALUES;
				continue;
			}

			if (args[i].equals("-l")){
				parser_state=CODE_LINES;
				continue;
			}

			switch (parser_state){
			case HELP: break;
			case FILE_NAME: out=new PrintStream(new FileOutputStream(args[i]));
				break;
			case CODE_LINES:
				try{
					lines_to_do=Integer.parseInt(args[i]);
				}catch (NumberFormatException e){
					System.out.println("-l: Integer number of lines expected");
					parser_state=HELP;
				}
				break;

			case SOURCE_VALUES:
				r=get_rc_by_name(args[i++]);
				if (i>=args.length){
					parser_state=HELP;
				}else{
				  try{
					r.value=Integer.parseInt(args[i]);
				  }catch (NumberFormatException e){
					System.out.println("-s: Integer number for the resource starting value expected: '"+r.name+"'");
					parser_state=HELP;
				  }
				}
				break;

			case DESTINATION_VALUES:
				r=get_rc_by_name(args[i++]);
				if (i>=args.length){
					parser_state=HELP;
				}else{
				  try{
					new_target(r, Integer.parseInt(args[i]));
				  }catch (NumberFormatException e){
					System.out.println("-d: Integer number for the resource destination value expected: '"+r.name+"'");
					parser_state=HELP;
				  }
				}
				break;
			}

			if (parser_state==HELP){
				break;
			}
		}

		if (parser_state==HELP){
			System.out.println(helpscreen);
			System.exit(0);
		}

		//new_target(ax, 0);
		//new_target(bx, 1);

		int l=lines_to_do;
		generate();

		System.out.println("done\nlines requested: "+l+"\nlines generated: "+(l-lines_to_do));
	   }catch (PolyException ex){
		System.out.println(ex.getMessage());
	   }catch (Exception ex){
		ex.printStackTrace();		
	   }
	}

	public static Target new_target(Rc dest, int what) throws PolyException{
		return new Target(new MoveRc(dest, new Const(what), null));
	}

	public static void generate() throws PolyException{
		for(int k=1000; k-->0;){	// just fear of hanging
			int i,j, o;
			for (i=0, j=0, lines_can_do=0, o=0; i<tg_count; i++){
				if (targets[i].can_exec()){
					j++;
				}

				lines_can_do+=targets[i].min_lines();
				if (targets[i].stopped==Target.TERMINATED){
					o++;
				}
			}

			if (j==0){
				if (o!=tg_count){
					throw new PolyException (":( deadlock encountered... face the thing that should not be: "+(tg_count-o)+" threads left");
				}
				break;
			}

			force_endurance = (tg_count-o)<=1; // if less than two targets are unterminated, the last target should endure to produce the needed output

			// count resources, so poor distorters won't have to do that
			for (i=0, rc_read=0, rc_write=0; i<rc_count; i++){
				if (resources[i].can_read()){
					rc_read++;
				}
				if (resources[i].can_write()){
					rc_write++;
				}
			}

			j=rnd.nextInt(j);
			for (i=0; ; i++){
				if (targets[i].can_exec()){
					if (j--==0) break;
				}
			}

			targets[i].exec();
		}
	}

	public static Rc get_rc_read(int n){
	// get n-th resource, accessible for reading
		for (int i=0; i<rc_count; i++){
			if (resources[i].can_read()){
				if (n--==0){
					return resources[i];
				}
			}
		}

		return null;
	}

	public static Rc get_rc_write(int n){
	// get n-th resource, accessible for writing

		for (int i=0; i<rc_count; i++){
			if (resources[i].can_write()){
				if (n--==0){
					return resources[i];
				}
			}
		}

		return null;
	}

	public static Rc get_random_read(){
	// get a random source
		return get_rc_read(rnd.nextInt(rc_read));
	}

	public static Rc get_random_write(){
	// get a random destination
		return get_rc_write(rnd.nextInt(rc_write));
	}

	public static Rc get_rc_by_name(String n) throws PolyException{
		for (int i=0; i<rc_count; i++){
			if (resources[i].name.equals(n)){
				return resources[i];
			}
		}

		return new Register(n, 0);
	}
}
