/*
 * this is the most primitive signature searcher.
 * i wanted to introduce simple language into signatures, which would allow
 * me calculate complex offsets and check for encrypted signatures,
 * but we are not designing The Ultimate Av, but merely a simple example
 * to show decent people how easily it can be done, and that it's
 * even more easy to implement plug-ins and add-ons, than some might think
 * at first.
 */

package utils.tools;

import java.util.Vector;
import utils.av.*;
import utils.tools.signaturesearch.*;
import utils.tools.patchhealer.ResultPatch;

public class SignatureSearch extends ZemskijFershal {
	// VB syntax tokens
	static final String PREPENDER_ID = "prepender";
	static final String TRAILER_ID = "trailer";
	static final String NAME_ID = "name";
	static final String SHORTER_ID = "shorter";
	static final String SIGNATURE_ID = "signature";
	static final String BYTES_ID = "bytes";
	static final String OFFSET_ID = "offset";
	static final String LEN_ID = "length";
	static final String ORIGIN_ID = "origin";

	static final String ENDOFENTRY_ID = ".";
	static final String ASSIGN_ID = "=";

	static final String PATCH_HEALER = "PatchHealer";

	static final String NEXT_HEURISTIC = "Belkodav"; // VCG detector


	Vector virusEntries = new Vector();

	public SignatureSearch() throws AVException, Exception{
		// read environment to determine Signature set
		VirBase vb = new VirBase(environment);
		String word;
		VirusEntry ve;

		while (!vb.eof()){
			word = vb.read_word();

			if(word.equals(PREPENDER_ID)){
				ve = new Prepender();
			}else{
				if (word.equals(TRAILER_ID)){
					ve = new Trailer();
				}else{
					throw new Exception("unknown entry in virus base: "+word);
				}
			}

			ve.virusHeadOffset = vb.read_int();
			word = vb.read_word();
			while (!word.equals(ENDOFENTRY_ID)){ // end of entry
				if (word.equals(NAME_ID)){
					vb.assert(ASSIGN_ID);

					ve.name = vb.read_word();
				}else{
				  if (word.equals(SIGNATURE_ID)){
				    Signature sig = new Signature();
				    sig.offset = vb.read_int();

				    vb.assert(ASSIGN_ID);
				
				    sig.signature = vb.read_bytes();

				    ve.signatures.add(sig);
				  }else{
				    if (word.equals(SHORTER_ID)){
					vb.assert(ASSIGN_ID);
					ve.shorten_by = vb.read_int();
				    }else{

				    // insert "if" clause here, when VB syntax extends
				      vb.assert(word, BYTES_ID);

				      Bytes b = new Bytes();
				      word = vb.read_word();
				      while (!word.equals(ENDOFENTRY_ID)){
					// yet all fields have '=' sign
					vb.assert(ASSIGN_ID);

					if (word.equals(OFFSET_ID)){
					  b.offset = vb.read_int();
					}else{
					  if (word.equals(ORIGIN_ID)){
					    b.origin = vb.read_int();
					  }else{
					    vb.assert(word, LEN_ID);
					    b.len = vb.read_int();
					  }
					}

					if (vb.eof()){
						vb.assert(ENDOFENTRY_ID);
					}
					word=vb.read_word();
				      }

				      ve.bytes.add(b);
				    }
				  }
				}

				if (vb.eof()){
					vb.assert(ENDOFENTRY_ID);
				}
				word = vb.read_word();
			}

			virusEntries.add(ve);
			vb.skip_spaces();
		}

	}

	public Result apply( Result prev ){
		prev.consult=NEXT_HEURISTIC;
					// here we might like to put name of
					// "Ninnyish based" detector.
				// who may use the same PatchHealer class for healing.

		VirusEntry ve;
		Signature sig;
next_stamm:
		for (int i=0; i<virusEntries.size(); i++){
			ve = (VirusEntry)(virusEntries.get(i));

			int headOffset = ve.vir_beg(prev.file);
			for (int j=0; j<ve.signatures.size(); j++){
				sig = (Signature)(ve.signatures.get(j));
				for (int k=0, o=0; k<sig.signature.length; k++){
					o=headOffset+k+sig.offset;
					if (o<0 || o>=prev.file.length || 
					    sig.signature[k]!=prev.file[o]){
						continue next_stamm;
					}
				}
			}

			// of course, we could do the work of restoring bytes
			// right here.
			// but let us complicate this task to show people
			// pluses of the moduled approach

			int [] patches = new int [3*ve.bytes.size()];
			Bytes b;
			for (int j=0; j<ve.bytes.size(); j++){
				b=(Bytes)(ve.bytes.get(j));
				patches[j*3]=b.offset;
				patches[j*3+1]=b.origin+headOffset;
				patches[j*3+2]=(b.len>0)? b.len : (prev.file.length-b.origin-headOffset);
			}

			return new ResultPatch(ve.name, PATCH_HEALER, Result.DETECTED, prev.file, prev.save, prev, prev.file.length-ve.shorten_by, patches);
		};

		return prev;
	}

}


